/* 
 * FILE --- rbf.hh 
 *
 * PURPOSE --- Header file for a class that implements a Radial Basis 
 * Function (RBF) neural network (NN) with gaussian nodes. 
 * The user specifies the number of inputs, the number of outputs and the 
 * number of the neurons (i.e., the gaussian units). The user needs to provide
 * the centers and the spreads og the gaussian nodes. Moreover, s/he can either 
 * specify the output parameters or let them be choosen randomly within the user defined range.
 * The utility has a build-in adaptation capability of the output parameters
 * based on the Normalized Gradient Method or the so called Projection Algorithm.
 * Therefore it can be used in both direct and indirect adaptive control schemes as
 * a function approximator.
 *
 * This utility uses the matrix tool for some matrix maltiplications.
 */


// Prevent Multiple Inclusion
#ifndef RBF_HH
#define RBF_HH


// Include Files
#include <stdlib.h>
#include "matrix.hh"


// Local Definitions and Macros
#define RBF_NO_ERROR 0
#define RBF_MEM_ERROR 1

class RBF
{
public:
  
  // Constructor 
  RBF(int _num_of_inputs, int _num_of_outputs, int _num_of_neurons);

  // Destructor
  ~RBF();

  // Functions to set the spreads of the neurons
  void set_spreads(double _spreads);
  void set_spreads(double *_spreads);
  void set_spreads(double **_spreads);

  // Functions to set the centers of the neurons
  void set_centers(double _centers);
  void set_centers(double **_centers);

  // Functions to set the output parameters
  void set_output_params(double _out_params);
  void set_output_params(double _low, double _high);
  void set_output_params(const double *_out_params);
  void set_output_params(const double **_out_params);

  // Functions to set the centers of the neurons
  void set_eta(double _eta) { eta = _eta; };
  double get_eta(void) { return eta; };

  // The function that given the inputs computes the output of the RBF. 
  // input = an array of inputs of size _num_of_inputs
  // output = an array of outputs of size _num_of_outputs
  void control(const double *input, double *output);

  // The function that given the error updates the output parameters 
  void update_output_params(const double *error);
  void update_output_params(const double *error, double norm2);
  void update_output_params(const double *error, double norm2, double uu);
  void bound_output_params(double _low, double _high);

  // Variable to hold the norm of phi (needed for indirect adaptive control)
  double phi_norm; 

  int rbf_err; // Variable to hold the error condition

private:
  // Allocate/free memory for the arrays
  void allocate_memory(void);
  void free_memory(void);

  // Function that calculates the activation of each neuron
  void get_activation(const double *input);
  // Function that transfers the outputs from a matrix to the output variable
  void get_output(double *_y);

  int num_of_inputs;  // # of inputs (to be specified by the user)
  int num_of_outputs;  // # of outputs (to be specified by the user)
  int num_of_neurons;  // # of neurons (to be specified by the user)

  // The the vectros for the parametrs
  double **centers;  // An array of centers (each center is a vector)
  double **spreads;   // An array of spreads (each spread is a vector)
  matrix *theta;     // An array of output parameters
  matrix *phi;       // An array of function values of the neurons

  double eta; // Adaptation gain

  // Conveniance variables
  matrix *y; // Output vector
  matrix *e; // Error vector (needed for adaptive control)
};

#endif // RBF_HH
